<?xml version="1.0" encoding="utf-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:xs="http://www.w3.org/2001/XMLSchema"
                xmlns:utf8="http://www.thoughtcrime.us/ns/xslt-utf8-decode"
                version="2.0">
  
  <!-- Decodes a UTF-8 string expressed as a sequence of octets. -->
  <xsl:function name="utf8:decode-string" as="xs:string">
    <xsl:param name="octets" as="xs:integer*"/>
    <!-- We use xsl:for-each-group because I can't think of any other
	 way to process a UTF-8 octet string.  I tried doing it
	 recursively, but I hit the stack limit, and I don't think
	 XSLT/XPath prescribe tail-call optimization. -->
    <!-- Guess what?!  xsl:for-each-group cannot operate on ANYTHING
	 BUT XML when group-starting-with is used!  Argh! -->
    <xsl:variable name="octets-xml" as="element(utf8:octet)*">
      <xsl:for-each select="$octets">
	<utf8:octet val="{.}"/>
      </xsl:for-each>
    </xsl:variable>
    <xsl:variable name="codepoints" as="xs:integer*">
      <xsl:for-each-group select="$octets-xml"
			  group-starting-with="utf8:octet[(0 = (@val idiv 128)) or
					                  (6 = (@val idiv 32)) or
							  (14 = (@val idiv 16)) or
							  (30 = (@val idiv 8))]">
	<xsl:sequence select="utf8:decode-character (current-group ()/@val)"/>
      </xsl:for-each-group>
    </xsl:variable>
    <xsl:sequence select="codepoints-to-string ($codepoints)"/>
  </xsl:function>
  
  <!-- Decodes a single UTF-8 character sequence, as octets, into a
       sequence of Unicode codepoints.  There must be at least one
       input octet.  The first octet, and only the first, must be the
       beginning of a UTF-8 character sequence (if only of length 1).
       Extra characters after the sequence are assumed to be illegal,
       non-sequence-starting octets, and are each replaced with the
       U+FFFD REPLACEMENT CHARACTER.
       
       Illegal sequences, sequences that decode to UTF-16 surrogate
       words, and sequences that decode to characters that cannot be
       expressed in XML are all replaced with U+FFFD. -->
  <xsl:function name="utf8:decode-character" as="xs:integer+">
    <xsl:param name="octets" as="xs:integer+"/>
    <xsl:variable name="seq-len" as="xs:integer?"
		  select="if (0 &lt;= $octets[1] and $octets[1] &lt; 128) then 1 else
			  if (192 &lt;= $octets[1] and $octets[1] &lt; 224) then 2 else
			  if (224 &lt;= $octets[1] and $octets[1] &lt; 240) then 3 else
			  if (240 &lt;= $octets[1] and $octets[1] &lt; 248) then 4 else ()"/>
    <xsl:variable name="octets-consumed" as="xs:integer"
		  select="if (exists ($seq-len)) then $seq-len else 1"/>
    <xsl:variable name="is-not-truncated-seq" as="xs:boolean"
		  select="exists ($seq-len) and
			  ($seq-len &lt;= count ($octets))"/>
    <!-- XPath doesn't support bitwise operations.  Who knew?!  It
	 doesn't support integer expression in binary or hex either.
	 Who knew?!
	 
	 We can cope with x+y in lieu of x|y and x*2**n for x<<n.
	 
	 Oh, wait.  XPath doesn't support exponentiation.  Who knew?!
	 -->
    <xsl:variable name="decoded-codepoint" as="xs:integer?">
      <xsl:choose>
	<xsl:when test="not ($is-not-truncated-seq)"/>
	<xsl:when test="$seq-len = 1">
	  <xsl:sequence select="$octets[1]"/>
	</xsl:when>
	<xsl:when test="$seq-len = 2">
	  <xsl:sequence select="(($octets[1] mod 32) * 64) +
				$octets[2] mod 64"/>
	</xsl:when>
	<xsl:when test="$seq-len = 3">
	  <xsl:sequence select="(($octets[1] mod 16) * 4096) +
				(($octets[2] mod 64) * 64) +
				$octets[3] mod 64"/>
	</xsl:when>
	<xsl:when test="$seq-len = 4">
	  <xsl:sequence select="(($octets[1] mod 8) * 262144) +
				(($octets[2] mod 64) * 4096) +
				(($octets[3] mod 64) * 64) +
				$octets[4] mod 64"/>
	</xsl:when>
      </xsl:choose>
    </xsl:variable>
    <!-- Verify that we haven't been tricked into decoding an overly
	 long sequence. -->
    <xsl:variable name="is-correct-seq-len" as="xs:boolean">
      <xsl:choose>
	<xsl:when test="exists ($decoded-codepoint)">
	  <xsl:variable name="correct-seq-len" as="xs:integer?"
			select="if (0 &lt;= $decoded-codepoint and
				    $decoded-codepoint &lt; 128) then 1 else
				if (128 &lt;= $decoded-codepoint and
				    $decoded-codepoint &lt; 2048) then 2 else
				if (2048 &lt;= $decoded-codepoint and
				    $decoded-codepoint &lt; 65536) then 3 else
				if (65536 &lt;= $decoded-codepoint and
				    $decoded-codepoint &lt; 1114112) then 4 else ()"/>
	  <xsl:sequence select="$seq-len = $correct-seq-len"/>
	</xsl:when>
	<xsl:otherwise>
	  <xsl:sequence select="false ()"/>
	</xsl:otherwise>
      </xsl:choose>
    </xsl:variable>
    <!-- Verify that we haven't decoded a codepoint reserved for
	 UTF-16 surrogate words. -->
    <xsl:variable name="is-not-surrogate" as="xs:boolean"
		  select="exists ($decoded-codepoint) and
			  not (55296 &lt;= $decoded-codepoint and
			       $decoded-codepoint &lt; 57344)"/>
    <!-- c.f. http://www.w3.org/TR/REC-xml/#charsets -->
    <xsl:variable name="is-valid-xml-char" as="xs:boolean"
		  select="exists ($decoded-codepoint) and
			  (($decoded-codepoint = 9) or
			   ($decoded-codepoint = 10) or
			   ($decoded-codepoint = 13) or
			   (32 &lt;= $decoded-codepoint and
			    $decoded-codepoint &lt; 55296) or
			   (57344 &lt;= $decoded-codepoint and
			    $decoded-codepoint &lt; 65534) or
			   (65536 &lt;= $decoded-codepoint and
			    $decoded-codepoint &lt; 1114112))"/>
    <xsl:if test="not ($is-not-truncated-seq and $is-correct-seq-len and
		       $is-not-surrogate and $is-valid-xml-char)">
      <xsl:message terminate="yes">
	<xsl:text>Processing seq: </xsl:text>
	<xsl:value-of select="string-join (for $octet in $octets return xs:string ($octet), ', ')"/>
	<xsl:text>&#10;</xsl:text>
	<xsl:value-of select="concat ('seq-len: ', $seq-len, '&#10;')"/>
	<xsl:value-of select="concat ('octets-consumed: ', $octets-consumed, '&#10;')"/>
	<xsl:value-of select="concat ('is-not-truncated-seq: ', $is-not-truncated-seq, '&#10;')"/>
	<xsl:value-of select="concat ('is-correct-seq-len: ', $is-correct-seq-len, '&#10;')"/>
	<xsl:value-of select="concat ('is-not-surrogate: ', $is-not-surrogate, '&#10;')"/>
	<xsl:value-of select="concat ('is-valid-xml-char: ', $is-valid-xml-char, '&#10;')"/>
      </xsl:message>
    </xsl:if>
    <!-- 65533 is U+FFFD REPLACEMENT CHARACTER -->
    <xsl:sequence select="if ($is-not-truncated-seq and $is-correct-seq-len and
			      $is-not-surrogate and $is-valid-xml-char)
			  then $decoded-codepoint else 65533"/>
    <!-- Replace unconsumed octets with U+FFFD. -->
    <xsl:sequence select="for $octet in subsequence ($octets, 1 + $octets-consumed)
			  return 65533"/>
  </xsl:function>
</xsl:stylesheet>
